 /*
 * Connections          - Arduino Nano/Micro with Atmega 328P
 * ----------------------------------------------------------
 * 
 * Receiver
 * Ext - (brown)        - GND
 * Ext + (red)          - Vin/RAW
 * Ext Sig (orange)     - TX - resistor 5.1 kOhm - RX (resistor is only needed when jeti box key inputs are used)
 * 
 * GY-63/MS5611 (cover sensor with black heat shrink tube as it is very sensitive to (sun) light)
 * 1 VCC                - 5V/VCC
 * 2 GND                - GND
 * 3 SCL                - A5
 * 4 SDA                - A4
 * 
 * Voltage divider rx battery 1
 * battery - (black)    - GND
 * battery + (red)      - resistor 51kOhm - A0 - resistor 51kOhm - GND (voltage divider max 10V -> 5V)
 * 
 * Voltage divider rx battery 2
 * battery - (black)    - GND
 * battery + (red)      - resistor 51kOhm - A1 - resistor 51kOhm - GND (voltage divider max 10V -> 5V)
 * 
 * Voltage divider drive battery
 * battery - (black)    - GND
 * battery + (red)      - resistor 510kOhm - A2 - resistor 51kOhm - GND (voltage divider max 55V -> 5V, max 12s Lipo)
 * 
 * ACS758 current sensor
 * 1 VCC                - 5V/VCC - capacitor 0.1uF - GND (place capacitor close to current sensor for stability)
 * 2 GND                - GND
 * 3 VIOUT              - resistor 100kOhm - A3 - capacitor 10nF - GND (lowpass filter)
 * 
 * MPXV7002 pitot tube pressure sensor (always cover the pitot tube before start up)
 * 1 GND                - GND
 * 2 VS                 - 5V/VCC
 * 3 VOUT               - A6
 */

#define SAMPLING_PERIOD 0.013 // s
#define SAMPLES_ALTITUDE 100 // 100: good compromise between reaction time and noise level of vario, for faster reaction decrease number, to reduce noise increase number
#define FILTER_CUTOFF_FREQUENCY 3 // Hz
#define MS5611_ADDRESS 0x77 // 0x76 or 0x77

#define RX_VOLTAGE_1_PIN A0
#define RX_VOLTAGE_2_PIN A1
#define DRIVE_VOLTAGE_PIN A2
#define DRIVE_CURRENT_PIN A3
#define PITOT_PRESSURE_PIN A6
#define LED_PIN 13

#include "JetiExProtocol.h"
#include <Wire.h>
#include <EEPROM.h>

JetiExProtocol jetiEx;

unsigned long timeSample; // us
float airTemperature, airTemperature0; // Kelvin
float airPressure, airPressure0; // Pa
float airDensity; // kg/m^3
float altitude; // m
float altitudeBuffer[SAMPLES_ALTITUDE]; // m
float climbRate; // m/s
float rxVoltage1; // V
float rxVoltage2; // V
float driveVoltage; // V
float driveCurrent; // A
float drivePower; // W
float driveCapacity; // Ah
float pitotPressure; // Pa
float airspeedSquared; // m^2/s^2
float airspeed; // m/s
long samplingPeriodMargin; // us

struct settings {
  float rxVoltage1Scale;
  float rxVoltage2Scale;
  float driveVoltageScale;
  float driveCurrentOffset;
  float driveCurrentScale;
  float pitotPressureOffset;
  float pitotPressureScale;
  float airspeedRatio;
  bool totalEnergieCompensation;
} set;

JETISENSOR_CONST sensors[] PROGMEM = {
  //id  name              unit      data type               precision 0->0, 1->0.0, 2->0.00
  { 1,  "Rel. Altit",     "m",      JetiSensor::TYPE_14b,   0},
  { 2,  "Climb Rate",     "m/s",    JetiSensor::TYPE_14b,   2},
  { 3,  "RX Voltage 1",   "V",      JetiSensor::TYPE_14b,   2},
  { 4,  "RX Voltage 2",   "V",      JetiSensor::TYPE_14b,   2},
  { 5,  "Drive Voltage",  "V",      JetiSensor::TYPE_14b,   2},
  { 6,  "Drive Current",  "A",      JetiSensor::TYPE_14b,   1},
  { 7,  "Drive Power",    "W",      JetiSensor::TYPE_14b,   0},
  { 8,  "Drive Capacity", "Ah",     JetiSensor::TYPE_14b,   2},
  { 9,  "Airspeed",       "m/s",    JetiSensor::TYPE_14b,   2},
  //{10,  "Period Margin",  "us",     JetiSensor::TYPE_14b,   0},
  0 // end of array
};

void setup() {
  pinMode(LED_PIN, OUTPUT);
  digitalWrite(LED_PIN, HIGH);
  
  EEPROM.get(0, set);

  jetiEx.Start("LD Sensor E", sensors);
  showJetiBoxMenu(0);
  
  set.pitotPressureOffset = analogReadOversampling(PITOT_PRESSURE_PIN, 64);  set.totalEnergieCompensation = true;
  initAltitude();
}

void loop() {
  timeSample += SAMPLING_PERIOD * 1e6;
  samplingPeriodMargin = timeSample - micros();
  digitalWrite(LED_PIN, samplingPeriodMargin < 0); // switch led on if sampling period is too short
  while (long(timeSample - micros()) > 0); // wait for next sample

  MS5611GetTemperaturePressure(MS5611_ADDRESS, airTemperature, airPressure);
  altitude = airTemperature0 / 0.0065 * (1 - pow(airPressure / airPressure0, 1 / 5.255)); // m, Source: Wikipedia Deutsch - Barometrische Höhenformel - Typische Temperaturgradienten
  airDensity = airPressure / (287.058 * airTemperature); // kg/m^3, Source: Wikipedia Deutsch - Luftdichte - Höhenabhängigkeit

  pitotPressure = set.pitotPressureScale * abs(analogReadOversampling(PITOT_PRESSURE_PIN, 16) - set.pitotPressureOffset);
  if (pitotPressure < 0) {
    airspeedSquared = 0;
    airspeed = 0;
  } else {
    airspeedSquared = set.airspeedRatio * pitotPressure / airDensity; // m^2/s^2
    airspeed = sqrt(airspeedSquared); // m/s
  }
  
  float x = -0.5 * (SAMPLES_ALTITUDE - 1), sxx = 0, sxy = 0; // init linear regression of altitude samples
  for (int i = 0;  i < SAMPLES_ALTITUDE - 1; i++) { // build sums over all altitude samples
    altitudeBuffer[i] = altitudeBuffer[i + 1]; // shift altitude data buffer by one sample
    sxx += x * x;
    sxy += x * altitudeBuffer[i];
    x++;
  }
  altitudeBuffer[SAMPLES_ALTITUDE - 1] = altitude;
  if (set.totalEnergieCompensation) altitudeBuffer[SAMPLES_ALTITUDE - 1] += airspeedSquared / (2 * 9.81); // altitude which could be reached by pulling up the plane (all kinetic energy is converted into potential energy)
  sxx += x * x;
  sxy += x * altitudeBuffer[SAMPLES_ALTITUDE - 1];
  
  climbRate = sxy / (sxx * SAMPLING_PERIOD); // calculate climb rate as slope of linear regression

  filter(rxVoltage1, set.rxVoltage1Scale * analogRead(RX_VOLTAGE_1_PIN)); // V
  filter(rxVoltage2, set.rxVoltage2Scale * analogRead(RX_VOLTAGE_2_PIN)); // V
  filter(driveVoltage, set.driveVoltageScale * analogRead(DRIVE_VOLTAGE_PIN)); // V  
  filter(driveCurrent, set.driveCurrentScale * (analogRead(DRIVE_CURRENT_PIN) - set.driveCurrentOffset)); // A
  drivePower = driveVoltage * driveCurrent; // W
  driveCapacity += driveCurrent * SAMPLING_PERIOD / 3600; // Ah

  handleJetiBox();
  
  jetiEx.SetSensorValue(1, altitude);
  jetiEx.SetSensorValue(2, climbRate * 100);
  jetiEx.SetSensorValue(3, rxVoltage1 * 100);
  jetiEx.SetSensorValue(4, rxVoltage2 * 100);
  jetiEx.SetSensorValue(5, driveVoltage * 10);
  jetiEx.SetSensorValue(6, driveCurrent * 10);
  jetiEx.SetSensorValue(7, drivePower);
  jetiEx.SetSensorValue(8, driveCapacity * 100);
  jetiEx.SetSensorValue(9, airspeed * 100);
  //jetiEx.SetSensorValue(10, constrain(samplingPeriodMargin, -8000, 8000));
  jetiEx.DoJetiSend();
}

float analogReadOversampling(int pin, int samples) {
  float sum = 0;
  for (int i = 0;  i < samples; i++) sum += analogRead(pin);
  return sum / samples;  
}

void filter(float &filteredValue, float newSample) {
  static float alpha = 2 * PI * SAMPLING_PERIOD * FILTER_CUTOFF_FREQUENCY / (2 * PI * SAMPLING_PERIOD * FILTER_CUTOFF_FREQUENCY + 1);
  static float beta = 1 - alpha;
  filteredValue = alpha * newSample + beta * filteredValue;
}

void initAltitude() {
  airTemperature0 = 0;
  airPressure0 = 0;
  for (int i = 0;  i < 64; i++) {
    MS5611GetTemperaturePressure(MS5611_ADDRESS, airTemperature, airPressure);
    airTemperature0 += airTemperature;
    airPressure0 += airPressure;
  }
  airTemperature0 /= 64;
  airPressure0 /= 64;
  timeSample = micros();
}
